from abc import ABC, abstractmethod
from copy import deepcopy
from functools import singledispatch, update_wrapper
from typing import TypeVar, Type, Mapping, Any, Iterable, Callable, Union, AbstractSet

from merakicommons.cache import lazy_property

from .common import PipelineContext, UnsupportedError, NotFoundError, TYPE_WILDCARD

T = TypeVar("T")


class DataSource(ABC):
    @staticmethod
    def unsupported(type: Type[T]) -> UnsupportedError:
        return UnsupportedError("The type \"{type}\" is not supported by this DataSource!".format(type=type.__name__))

    @lazy_property
    def provides(self):  # type: Union[Iterable[Type[T]], Type[Any]]
        """The types of objects the data store provides."""
        types = set()
        any_dispatch = False
        try:
            types.update(getattr(self.__class__, "get")._provides)
            any_dispatch = True
        except AttributeError:
            pass
        try:
            types.update(getattr(self.__class__, "get_many")._provides)
            any_dispatch = True
        except AttributeError:
            pass
        return types if any_dispatch else TYPE_WILDCARD

    @abstractmethod
    def get(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> T:
        """Gets a query from the data source.

        Args:
            query: The query being requested.
            context: The context for the extraction (mutable).

        Returns:
            The requested object.
        """
        pass

    @abstractmethod
    def get_many(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        """Gets a query from the data source, which contains a request for multiple objects.

        Args:
            query: The query being requested (contains a request for multiple objects).
            context: The context for the extraction (mutable).

        Returns:
            The requested objects.
        """
        pass

    @staticmethod
    def dispatch(method: Callable[[Any, Type[T], Mapping[str, Any], PipelineContext], Any]) -> Callable[[Any, Type[T], Mapping[str, Any], PipelineContext], Any]:
        dispatcher = singledispatch(method)
        provides = set()

        def wrapper(self: Any, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> Any:
            call = dispatcher.dispatch(type)
            try:
                return call(self, query, context=context)
            except TypeError:
                raise DataSource.unsupported(type)

        def register(type: Type[T]) -> Callable[[Any, Type[T], Mapping[str, Any], PipelineContext], Any]:
            provides.add(type)
            return dispatcher.register(type)

        wrapper.register = register
        wrapper._provides = provides
        update_wrapper(wrapper, method)
        return wrapper


class CompositeDataSource(DataSource):
    def __init__(self, sources: Iterable[DataSource]) -> None:
        self._sources = {}
        for source in sources:
            for provided_type in source.provides:
                try:
                    providing_sources = self._sources[provided_type]
                except KeyError:
                    providing_sources = []
                    self._sources[provided_type] = providing_sources
                providing_sources.append(source)

    @property
    def provides(self) -> AbstractSet[Type]:
        return self._sources.keys()

    def get_many(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        try:
            sources = self._sources[type]
        except KeyError as error:
            raise DataSource.unsupported(type) from error

        for source in sources:
            try:
                return source.get_many(type, deepcopy(query), context)
            except NotFoundError:
                continue
        raise NotFoundError()

    def get(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> T:
        try:
            sources = self._sources[type]
        except KeyError as error:
            raise DataSource.unsupported(type) from error

        for source in sources:
            try:
                return source.get(type, deepcopy(query), context)
            except NotFoundError:
                continue
        raise NotFoundError()



import pytest

from datapipelines import Query, QueryValidationError, QueryValidatorStructureError, validate_query


def test_has():
    valid = Query.has("test")

    with pytest.raises(QueryValidationError):
        valid({})

    with pytest.raises(QueryValidationError):
        valid({"dog": "cat"})

    assert valid({"test": "test"})
    assert valid({"test": 0})
    assert valid({"test": "test", "dog": "cat"})


def test_repeat_has():
    with pytest.raises(QueryValidatorStructureError):
        Query.has("test").has("dog")


def test_can_have():
    valid = Query.can_have("test")

    assert valid({})
    assert valid({"dog": "cat"})
    assert valid({"test": "test"})
    assert valid({"test": 0})
    assert valid({"test": "test", "dog": "cat"})


def test_repeat_can_have():
    with pytest.raises(QueryValidatorStructureError):
        Query.can_have("test").can_have("dog")


def test_repeat_have_can_have():
    with pytest.raises(QueryValidatorStructureError):
        Query.has("test").can_have("dog")

    with pytest.raises(QueryValidatorStructureError):
        Query.can_have("test").has("dog")


def test_has_as():
    valid = Query.has("test").as_(str)

    with pytest.raises(QueryValidationError):
        valid({})

    with pytest.raises(QueryValidationError):
        valid({"dog": "cat"})

    with pytest.raises(QueryValidationError):
        valid({"test": 0})

    assert valid({"test": "test"})
    assert valid({"test": "test", "dog": "cat"})


def test_can_have_as():
    valid = Query.can_have("test").as_(str)

    with pytest.raises(QueryValidationError):
        valid({"test": 0})

    assert valid({})
    assert valid({"dog": "cat"})
    assert valid({"test": "test"})
    assert valid({"test": "test", "dog": "cat"})


def test_repeat_as():
    with pytest.raises(QueryValidatorStructureError):
        Query.has("test").as_(str).as_(str)

    with pytest.raises(QueryValidatorStructureError):
        Query.can_have("test").as_(str).as_(str)


def test_has_as_any_of():
    valid = Query.has("test").as_any_of({str, int})

    with pytest.raises(QueryValidationError):
        valid({})

    with pytest.raises(QueryValidationError):
        valid({"dog": "cat"})

    with pytest.raises(QueryValidationError):
        valid({"test": 0.0})

    assert valid({"test": 0})
    assert valid({"test": "test"})
    assert valid({"test": "test", "dog": "cat"})


def test_can_have_as_any_of():
    valid = Query.can_have("test").as_any_of({str, int})

    with pytest.raises(QueryValidationError):
        valid({"test": 0.0})

    assert valid({})
    assert valid({"dog": "cat"})
    assert valid({"test": 0})
    assert valid({"test": "test"})
    assert valid({"test": "test", "dog": "cat"})


def test_repeat_as_any_of():
    with pytest.raises(QueryValidatorStructureError):
        Query.has("test").as_any_of({int, str}).as_any_of({int, str})

    with pytest.raises(QueryValidatorStructureError):
        Query.can_have("test").as_any_of({int, str}).as_any_of({int, str})


def test_repeat_as_as_any_of():
    with pytest.raises(QueryValidatorStructureError):
        Query.has("test").as_(str).as_any_of({int, str})

    with pytest.raises(QueryValidatorStructureError):
        Query.has("test").as_any_of({int, str}).as_(str)

    with pytest.raises(QueryValidatorStructureError):
        Query.can_have("test").as_(str).as_any_of({int, str})

    with pytest.raises(QueryValidatorStructureError):
        Query.can_have("test").as_any_of({int, str}).as_(str)


def test_has_or():
    valid = Query.has("test").or_("dog").or_("foo")

    with pytest.raises(QueryValidationError):
        valid({})

    assert valid({"test": "test"})
    assert valid({"dog": "cat"})
    assert valid({"foo": "bar"})
    assert valid({"test": 0})
    assert valid({"test": "test", "dog": "cat", "foo": "bar"})


def test_can_have_or():
    valid = Query.can_have("test").or_("dog").or_("foo")

    assert valid({})
    assert valid({"test": "test"})
    assert valid({"dog": "cat"})
    assert valid({"foo": "bar"})
    assert valid({"test": 0})
    assert valid({"test": "test", "dog": "cat", "foo": "bar"})


def test_has_as_or():
    valid = Query.has("test").as_(str).or_("dog").as_(int)

    with pytest.raises(QueryValidationError):
        valid({})

    with pytest.raises(QueryValidationError):
        valid({"test": 0})

    with pytest.raises(QueryValidationError):
        valid({"dog": "cat"})

    assert valid({"test": "test"})
    assert valid({"dog": 0})
    assert valid({"test": "test", "dog": 0})


def test_can_have_as_or():
    valid = Query.can_have("test").as_(str).or_("dog").as_(int)

    with pytest.raises(QueryValidationError):
        valid({"test": 0})

    with pytest.raises(QueryValidationError):
        valid({"dog": "cat"})

    assert valid({})
    assert valid({"test": "test"})
    assert valid({"dog": 0})
    assert valid({"test": "test", "dog": 0})


def test_has_as_any_of_or():
    valid = Query.has("test").as_any_of({str, int}).or_("dog").as_any_of({str, int})

    with pytest.raises(QueryValidationError):
        valid({})

    with pytest.raises(QueryValidationError):
        valid({"test": 0.0})

    with pytest.raises(QueryValidationError):
        valid({"dog": 0.0})

    assert valid({"test": "test"})
    assert valid({"test": 0})
    assert valid({"dog": "cat"})
    assert valid({"dog": 0})
    assert valid({"test": "test", "dog": 0})


def test_can_have_as_any_of_or():
    valid = Query.can_have("test").as_any_of({str, int}).or_("dog").as_any_of({str, int})

    with pytest.raises(QueryValidationError):
        valid({"test": 0.0})

    with pytest.raises(QueryValidationError):
        valid({"dog": 0.0})

    assert valid({})
    assert valid({"test": "test"})
    assert valid({"test": 0})
    assert valid({"dog": "cat"})
    assert valid({"dog": 0})
    assert valid({"test": "test", "dog": 0})


def test_and():
    valid = Query.has("test").and_("dog").and_("foo")

    with pytest.raises(QueryValidationError):
        valid({})

    with pytest.raises(QueryValidationError):
        valid({"test": "test"})

    with pytest.raises(QueryValidationError):
        valid({"dog": "cat"})

    with pytest.raises(QueryValidationError):
        valid({"foo": "bar"})

    with pytest.raises(QueryValidationError):
        valid({"test": 0})

    assert valid({"test": "test", "dog": "cat", "foo": "bar"})


def test_has_as_and():
    valid = Query.has("test").as_(str).and_("dog").as_(int)

    with pytest.raises(QueryValidationError):
        valid({})

    with pytest.raises(QueryValidationError):
        valid({"test": 0})

    with pytest.raises(QueryValidationError):
        valid({"dog": "cat"})

    with pytest.raises(QueryValidationError):
        valid({"test": "test"})

    with pytest.raises(QueryValidationError):
        valid({"dog": 0})

    assert valid({"test": "test", "dog": 0})


def test_can_have_as_and():
    valid = Query.can_have("test").as_(str).and_("dog").as_(int)

    with pytest.raises(QueryValidationError):
        valid({"test": 0})

    with pytest.raises(QueryValidationError):
        valid({"test": "test"})

    with pytest.raises(QueryValidationError):
        valid({"dog": "cat"})

    with pytest.raises(QueryValidationError):
        valid({"dog": 0})

    with pytest.raises(QueryValidationError):
        valid({"test": 0, "dog": "cat"})

    assert valid({})
    assert valid({"test": "test", "dog": 0})


def test_has_as_any_of_and():
    valid = Query.has("test").as_any_of({str, int}).and_("dog").as_any_of({str, int})

    with pytest.raises(QueryValidationError):
        valid({})

    with pytest.raises(QueryValidationError):
        valid({"test": 0.0})

    with pytest.raises(QueryValidationError):
        valid({"dog": 0.0})

    with pytest.raises(QueryValidationError):
        valid({"test": "test"})

    with pytest.raises(QueryValidationError):
        valid({"test": 0})

    with pytest.raises(QueryValidationError):
        valid({"dog": "cat"})

    with pytest.raises(QueryValidationError):
        valid({"dog": 0})

    assert valid({"test": "test", "dog": "cat"})
    assert valid({"test": "test", "dog": 0})
    assert valid({"test": 0, "dog": "cat"})
    assert valid({"test": 0, "dog": 0})


def test_can_have_as_any_of_and():
    valid = Query.can_have("test").as_any_of({str, int}).and_("dog").as_any_of({str, int})

    with pytest.raises(QueryValidationError):
        valid({"test": 0.0})

    with pytest.raises(QueryValidationError):
        valid({"test": 0})

    with pytest.raises(QueryValidationError):
        valid({"test": "test"})

    with pytest.raises(QueryValidationError):
        valid({"dog": 0.0})

    with pytest.raises(QueryValidationError):
        valid({"dog": 0})

    with pytest.raises(QueryValidationError):
        valid({"dog": "cat"})

    with pytest.raises(QueryValidationError):
        valid({"test": 0.0, "dog": 0.0})

    assert valid({})
    assert valid({"test": "test", "dog": "cat"})
    assert valid({"test": "test", "dog": 0})
    assert valid({"test": 0, "dog": "cat"})
    assert valid({"test": 0, "dog": 0})


def test_has_nested_and_or():
    valid = Query.has("test").and_("cat").or_("dog")

    with pytest.raises(QueryValidationError):
        valid({})

    with pytest.raises(QueryValidationError):
        valid({"test": "test"})

    with pytest.raises(QueryValidationError):
        valid({"dog": "cat"})

    with pytest.raises(QueryValidationError):
        valid({"cat": "dog"})

    with pytest.raises(QueryValidationError):
        valid({"test": "test", "foo": "bar"})

    assert valid({"test": "test", "cat": "dog"})
    assert valid({"test": "test", "dog": "cat"})


def test_has_nested_or_and():
    valid = Query.has("test").or_("cat").and_("dog")

    with pytest.raises(QueryValidationError):
        valid({})

    with pytest.raises(QueryValidationError):
        valid({"dog": "cat"})

    with pytest.raises(QueryValidationError):
        valid({"cat": "dog"})

    assert valid({"test": "test"})
    assert valid({"dog": "cat", "cat": "dog"})


def test_can_have_nested_and_or():
    valid = Query.can_have("test").and_("cat").or_("dog")

    with pytest.raises(QueryValidationError):
        valid({"test": "test"})

    with pytest.raises(QueryValidationError):
        valid({"cat": "dog"})

    with pytest.raises(QueryValidationError):
        valid({"dog": "cat"})

    assert valid({})
    assert valid({"test": "test", "dog": "cat"})
    assert valid({"test": "test", "cat": "dog"})
    assert valid({"test": "test", "dog": "cat", "cat": "dog"})


def test_can_have_nested_or_and():
    valid = Query.can_have("test").or_("cat").and_("dog")

    with pytest.raises(QueryValidationError):
        valid({"dog": "cat"})

    with pytest.raises(QueryValidationError):
        valid({"cat": "dog"})

    assert valid({})
    assert valid({"test": "test"})
    assert valid({"dog": "cat", "cat": "dog"})
    assert valid({"test": "test", "foo": "bar"})
    assert valid({"test": "test", "cat": "dog"})
    assert valid({"test": "test", "dog": "cat"})
    assert valid({"test": "test", "dog": "cat", "cat": "dog"})


def test_also():
    valid = Query.has("test").also.has("dog").also.has("foo")

    with pytest.raises(QueryValidationError):
        valid({})

    with pytest.raises(QueryValidationError):
        valid({"test": "test"})

    with pytest.raises(QueryValidationError):
        valid({"dog": "cat"})

    with pytest.raises(QueryValidationError):
        valid({"foo": "bar"})

    with pytest.raises(QueryValidationError):
        valid({"test": 0})

    assert valid({"test": "test", "dog": "cat", "foo": "bar"})


def test_repeat_also():
    with pytest.raises(QueryValidatorStructureError):
        Query.has("test").also.also


def test_default():
    valid = Query.can_have("test").with_default("test")

    query = {}
    assert valid(query)
    assert query == {"test": "test"}


def test_bad_default():
    with pytest.raises(QueryValidatorStructureError):
        Query.has("test").with_default("test")


def test_wrong_default_type():
    valid = Query.can_have("test").with_default("test")

    with pytest.raises(QueryValidationError):
        valid({"test": 1})


def test_no_default_type():
    valid = Query.can_have("test").with_default("test")

    query = {"test": "dog"}
    assert valid(query)
    assert query == {"test": "dog"}


def test_default_supplier():
    x = 0

    def supplier(query, context):
        nonlocal x
        x += 1
        return "test"

    valid = Query.can_have("test").with_default(supplier, str)

    query = {"test": "dog"}
    assert valid(query)
    assert query == {"test": "dog"}
    assert x == 0

    with pytest.raises(QueryValidationError):
        valid({"test": 1})
    assert x == 0

    query = {}
    assert valid(query)
    assert query == {"test": "test"}
    assert x == 1

    query = {}
    assert valid(query)
    assert query == {"test": "test"}
    assert x == 2


def test_validate_decorator():
    def pre_transform(query):
        if "test0" in query:
            query["test1"] = query["test0"]

    def pre_transform2(query):
        if "test1" in query:
            query["test2"] = int(query["test1"])

    validator = Query.has("test1").as_(str).also.has("test2").as_(int)

    @validate_query(validator, pre_transform, pre_transform2)
    def get(self, query, context=None):
        return query["test2"]

    with pytest.raises(QueryValidationError):
        get(None, {"cat": "dog"})

    with pytest.raises(ValueError):
        get(None, {"test1": "one"})

    assert get(None, {"test0": "1"}) == 1



from abc import ABC, abstractmethod
from functools import singledispatch, update_wrapper
from typing import TypeVar, Type, Any, Iterable, Callable, Union, AbstractSet

from .common import PipelineContext, UnsupportedError, TYPE_WILDCARD

T = TypeVar("T")


class DataSink(ABC):
    @staticmethod
    def unsupported(type: Type[T]) -> UnsupportedError:
        return UnsupportedError("The type \"{type}\" is not supported by this DataSink!".format(type=type.__name__))

    @property
    def accepts(self):  # type: Union[Iterable[Type[T]], Type[Any]]
        """The types of objects the data sink can store."""
        types = set()
        any_dispatch = False
        try:
            types.update(getattr(self.__class__, "put")._accepts)
            any_dispatch = True
        except AttributeError:
            pass
        try:
            types.update(getattr(self.__class__, "put_many")._accepts)
            any_dispatch = True
        except AttributeError:
            pass
        return types if any_dispatch else TYPE_WILDCARD

    @abstractmethod
    def put(self, type: Type[T], item: T, context: PipelineContext = None) -> None:
        """Puts an object into the data sink.

        Args:
            type: The type of the object being inserted.
            item: The object to be inserted.
            context: The context of the insertion (mutable).
        """
        pass

    @abstractmethod
    def put_many(self, type: Type[T], items: Iterable[T], context: PipelineContext = None) -> None:
        """Puts multiple objects of the same type into the data sink.

        Args:
            type: The type of the objects being inserted.
            items: The objects to be inserted.
            context: The context of the insertion (mutable).
        """
        pass

    @staticmethod
    def dispatch(method: Callable[[Any, Type[T], Any, PipelineContext], None]) -> Callable[[Any, Type[T], Any, PipelineContext], None]:
        dispatcher = singledispatch(method)
        accepts = set()

        def wrapper(self: Any, type: Type[T], items: Any, context: PipelineContext = None) -> None:
            call = dispatcher.dispatch(type)
            try:
                return call(self, items, context=context)
            except TypeError:
                raise DataSink.unsupported(type)

        def register(type: Type[T]) -> Callable[[Any, Type[T], Any, PipelineContext], None]:
            accepts.add(type)
            return dispatcher.register(type)

        wrapper.register = register
        wrapper._accepts = accepts
        update_wrapper(wrapper, method)
        return wrapper


class CompositeDataSink(DataSink):
    def __init__(self, sinks: Iterable[DataSink]) -> None:
        self._sinks = {}
        for sink in sinks:
            for accepted_type in sink.accepts:
                try:
                    accepting_sinks = self._sinks[accepted_type]
                except KeyError:
                    accepting_sinks = set()
                    self._sinks[accepted_type] = accepting_sinks
                accepting_sinks.add(sink)

    @property
    def accepts(self) -> AbstractSet[Type]:
        return self._sinks.keys()

    def put_many(self, type: Type[T], items: Iterable[T], context: PipelineContext = None) -> None:
        try:
            sinks = self._sinks[type]
        except KeyError as error:
            raise DataSink.unsupported(type) from error

        items = list(items)
        for sink in sinks:
            sink.put_many(type, items, context)

    def put(self, type: Type[T], item: T, context: PipelineContext = None) -> None:
        try:
            sinks = self._sinks[type]
        except KeyError as error:
            raise DataSink.unsupported(type) from error

        for sink in sinks:
            sink.put(type, item, context)



from typing import Type, TypeVar

import pytest

from datapipelines import DataTransformer, CompositeDataTransformer, PipelineContext

T = TypeVar("T")
F = TypeVar("F")


#############################################
# Create simple DataTransformer for testing #
#############################################

class SimpleDataTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    @transform.register(float, int)
    def float_to_int(self, value: float, context: PipelineContext = None) -> int:
        return int(value)

    @transform.register(int, float)
    def int_to_float(self, value: int, context: PipelineContext = None) -> float:
        return float(value)

    @transform.register(str, int)
    def str_to_int(self, value: str, context: PipelineContext = None) -> int:
        return int(value)

    @transform.register(str, float)
    def str_to_float(self, value: str, context: PipelineContext = None) -> float:
        return float(value)


class IntFloatTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    @transform.register(int, float)
    def int_to_float(self, value: int, context: PipelineContext = None) -> float:
        return float(value)


class FloatIntTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    @transform.register(float, int)
    def float_to_int(self, value: float, context: PipelineContext = None) -> int:
        return int(value)


########################
# Unsupported Function #
########################

def test_unsupported():
    from datapipelines import UnsupportedError

    unsupported = DataTransformer.unsupported(int, 0.0)
    assert type(unsupported) is UnsupportedError

    unsupported = DataTransformer.unsupported(float, 0)
    assert type(unsupported) is UnsupportedError

    unsupported = DataTransformer.unsupported(str, None)
    assert type(unsupported) is UnsupportedError


#######################
# Transforms Function #
#######################

def test_transforms():
    transformer = SimpleDataTransformer()

    assert transformer.transforms == {
        int: {float},
        float: {int},
        str: {int, float}
    }


######################
# Transform Function #
######################

def test_transform():
    transformer = SimpleDataTransformer()

    value = 0
    assert type(transformer.transform(float, value)) is float

    value = 0.0
    assert type(transformer.transform(int, value)) is int

    value = "0"
    assert type(transformer.transform(int, value)) is int
    assert type(transformer.transform(float, value)) is float


def test_transform_unsupported():
    from datapipelines import UnsupportedError
    transformer = SimpleDataTransformer()

    value = 0
    with pytest.raises(UnsupportedError):
        transformer.transform(str, value)


############################
# CompositeDataTransformer #
############################

def test_composite_transforms():
    transformer = CompositeDataTransformer({IntFloatTransformer(), FloatIntTransformer()})
    assert transformer.transforms == {
        int: {float},
        float: {int}
    }


def test_composite_transform():
    transformer = CompositeDataTransformer({IntFloatTransformer(), FloatIntTransformer()})

    value = 0
    assert type(transformer.transform(float, value)) is float

    value = 0.0
    assert type(transformer.transform(int, value)) is int


def test_composite_transform_unsupported():
    from datapipelines import UnsupportedError
    transformer = CompositeDataTransformer({IntFloatTransformer(), FloatIntTransformer()})

    value = 0
    with pytest.raises(UnsupportedError):
        transformer.transform(str, value)



from typing import Type, TypeVar, Sequence, Union, Callable, Any, List, Set, Generic, Mapping, Iterable, Tuple, Generator
from functools import partial
from itertools import tee
from logging import getLogger
from copy import deepcopy

from networkx import DiGraph, dijkstra_path, NetworkXNoPath

from .transformers import DataTransformer
from .common import PipelineContext, NotFoundError, TYPE_WILDCARD
from .sources import DataSource
from .sinks import DataSink

LOGGER = getLogger(__name__)

_SOURCES = "sources"
_SINKS = "sinks"
_TRANSFORMER = "transformer"

_MAX_TRANSFORM_COST = 10000000  # (つ͡°͜ʖ͡°)つ


def _build_type_graph(sources: Iterable[DataSource], sinks: Iterable[DataSink], transformers: Iterable[DataTransformer]) -> DiGraph:
    graph = DiGraph()

    for source in sources:
        # We ignore wildcard sources in the graph since they won't require a search to determine whether they can provide a type
        if TYPE_WILDCARD is source.provides:
            LOGGER.info("Ignoring \"{source}\" in type graph, as it provides the wildcard type".format(source=source))
            continue

        provides = source.provides  # type: Iterable[Type]
        for provided_type in provides:
            try:
                provided_type_sources = graph[provided_type][_SOURCES]
            except KeyError:
                provided_type_sources = set()

            provided_type_sources.add(source)
            graph.add_node(provided_type, sources=provided_type_sources)
        LOGGER.info("Added source \"{source}\" to type graph".format(source=source))

    for sink in sinks:
        # We ignore wildcard sinks in the graph since they won't require a search to determine whether they can accept a type
        if TYPE_WILDCARD is sink.accepts:
            LOGGER.info("Ignoring \"{sink}\" in type graph, as it accepts the wildcard type".format(sink=sink))
            continue
        else:
            accepts = sink.accepts  # type: Iterable[Type]

        for accepted_type in accepts:
            try:
                accepted_type_sinks = graph[accepted_type][_SINKS]
            except KeyError:
                accepted_type_sinks = set()

            accepted_type_sinks.add(sink)
            graph.add_node(accepted_type, sinks=accepted_type_sinks)
        LOGGER.info("Added sink \"{sink}\" to type graph".format(sink=sink))

    for transformer in transformers:
        # A wildcard transformer would be ridiculous so those don't exist
        for from_type, to_types in transformer.transforms.items():
            for to_type in to_types:
                # Add the type nodes into the graph if they aren't there already
                graph.add_node(from_type)
                graph.add_node(to_type)

                # Only the cheapest conversion between two types is stored. Replace it if this one is cheaper.
                try:
                    current_transformer = graph.adj[from_type][to_type][_TRANSFORMER]
                    cheapest_transformer = transformer if transformer.cost < current_transformer.cost else current_transformer
                except KeyError:
                    cheapest_transformer = transformer

                graph.add_edge(from_type, to_type, cost=transformer.cost, transformer=cheapest_transformer)
        LOGGER.info("Added transformer \"{transformer}\" to type graph".format(transformer=transformer))

    return graph


class NoConversionError(ValueError):
    pass


T = TypeVar("T")
S = TypeVar("S")


def _pairwise(iterable: Iterable[T]) -> Iterable[Tuple[T, T]]:
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def _identity(data: T, context: PipelineContext = None) -> T:
    return data


def _transform(transformer_chain: Sequence[Tuple[DataTransformer, Type]], data: S, context: PipelineContext = None) -> T:
    """Transform data to a new type.

    Args:
        transformer_chain: A sequence of (transformer, type) pairs to convert the data.
        data: The data to be transformed.
        context: The context of the transformations (mutable).

    Returns:
        The transformed data.
    """
    for transformer, target_type in transformer_chain:
        # noinspection PyTypeChecker
        data = transformer.transform(target_type, data, context)
    return data


class _SinkHandler(Generic[S, T]):
    def __init__(self, sink: DataSink, store_type: Type[S], transform: Callable[[T], S]) -> None:
        """Initializes a handler for a data sink.

        Args:
            sink: The data sink.
            store_type: ???
            transform: ???
        """
        self._sink = sink
        self._store_type = store_type
        self._transform = transform

    def put(self, item: T, context: PipelineContext = None) -> None:
        """Puts an objects into the data sink. The objects may be transformed into a new type for insertion if necessary.

        Args:
            item: The objects to be inserted into the data sink.
            context: The context of the insertion (mutable).
        """
        LOGGER.info("Converting item \"{item}\" for sink \"{sink}\"".format(item=item, sink=self._sink))
        item = self._transform(data=item, context=context)
        LOGGER.info("Puting item \"{item}\" into sink \"{sink}\"".format(item=item, sink=self._sink))
        self._sink.put(self._store_type, item, context)

    def put_many(self, items: Iterable[T], context: PipelineContext = None) -> None:
        """Puts multiple objects of the same type into the data sink. The objects may be transformed into a new type for insertion if necessary.

        Args:
            items: An iterable (e.g. list) of objects to be inserted into the data sink.
            context: The context of the insertions (mutable).
        """
        LOGGER.info("Creating transform generator for items \"{items}\" for sink \"{sink}\"".format(items=items, sink=self._sink))
        transform_generator = (self._transform(data=item, context=context) for item in items)
        LOGGER.info("Putting transform generator for items \"{items}\" into sink \"{sink}\"".format(items=items, sink=self._sink))
        self._sink.put_many(self._store_type, transform_generator, context)


class _SourceHandler(Generic[S, T]):
    def __init__(self, source: DataSource, source_type: Type[S], transform: Callable[[S], T], sinks: Mapping[_SinkHandler, bool]) -> None:
        """Initializes a handler for a data source.

        source: The data source.
        source_type: ???
        transform: ???
        sinks: ???
        """
        self._source = source
        self._source_type = source_type
        self._transform = transform
        self._before_transform = {sink for sink, do_transform in sinks.items() if not do_transform}
        self._after_transform = {sink for sink, do_transform in sinks.items() if do_transform}

    def get(self, query: Mapping[str, Any], context: PipelineContext = None) -> T:
        """Gets a query from the data source.

        1) Extracts the query from the data source.
        2) Inserts the result into any data sinks.
        3) Transforms the result into the requested type if it wasn't already.
        4) Inserts the transformed result into any data sinks.

        Args:
            query: The query being requested.
            context: The context for the extraction (mutable).

        Returns:
            The requested object.
        """
        result = self._source.get(self._source_type, deepcopy(query), context)
        LOGGER.info("Got result \"{result}\" from query \"{query}\" of source \"{source}\"".format(result=result, query=query, source=self._source))

        LOGGER.info("Sending result \"{result}\" to sinks before converting".format(result=result))
        for sink in self._before_transform:
            sink.put(result, context)

        LOGGER.info("Converting result \"{result}\" to request type".format(result=result))
        result = self._transform(data=result, context=context)

        LOGGER.info("Sending result \"{result}\" to sinks after converting".format(result=result))
        for sink in self._after_transform:
            sink.put(result, context)

        return result

    def _get_many_generator(self, result: Iterable[S], context: PipelineContext = None) -> Generator[T, None, None]:
        for item in result:
            LOGGER.info("Sending item \"{item}\" to sinks before converting".format(item=item))
            for sink in self._before_transform:
                sink.put(item, context)

            LOGGER.info("Converting item \"{item}\" to request type".format(item=item))
            item = self._transform(data=item, context=context)

            LOGGER.info("Sending item \"{item}\" to sinks after converting".format(item=item))
            for sink in self._after_transform:
                sink.put(item, context)

            yield item

    def get_many(self, query: Mapping[str, Any], context: PipelineContext = None, streaming: bool = False) -> Iterable[T]:
        """Gets a query from the data source, where the query contains multiple elements to be extracted.

        1) Extracts the query from the data source.
        2) Inserts the result into any data sinks.
        3) Transforms the results into the requested type if it wasn't already.
        4) Inserts the transformed result into any data sinks.

        Args:
            query: The query being requested.
            context: The context for the extraction (mutable).
            streaming: Specifies whether the results should be returned as a generator (default False).

        Returns:
            The requested objects or a generator of the objects if streaming is True.
        """
        result = self._source.get_many(self._source_type, deepcopy(query), context)
        LOGGER.info("Got results \"{result}\" from query \"{query}\" of source \"{source}\"".format(result=result, query=query, source=self._source))

        if not streaming:
            LOGGER.info("Non-streaming get_many request. Ensuring results \"{result}\" are a Iterable".format(result=result))
            result = list(result)

            LOGGER.info("Sending results \"{result}\" to sinks before converting".format(result=result))
            for sink in self._before_transform:
                sink.put_many(result, context)

            LOGGER.info("Converting results \"{result}\" to request type".format(result=result))
            result = [self._transform(data=item, context=context) for item in result]

            LOGGER.info("Sending results \"{result}\" to sinks after converting".format(result=result))
            for sink in self._after_transform:
                sink.put_many(result, context)

            return result
        else:
            LOGGER.info("Streaming get_many request. Returning result generator for results \"{result}\"".format(result=result))
            return self._get_many_generator(result)


class DataPipeline(object):
    def __init__(self, elements: Sequence[Union[DataSource, DataSink]], transformers: Iterable[DataTransformer] = None) -> None:
        """Initializes a data pipeline.

        Args:
            elements: The data stores and data sinks for this pipeline.
            transformers: The data transformers for this pipeline.
        """
        if not elements:
            raise ValueError("Elements must be a non-empty sequence of DataSources and DataSinks")

        if transformers is None:
            transformers = set()

        sources = set()  # type: Set[DataSource]
        sinks = set()  # type: Set[DataSink]
        targets = []  # type: List[Tuple[DataSource, Set[DataSink]]]
        for element in elements:
            if isinstance(element, DataSource):
                sources.add(element)
                targets.append((element, set(sinks)))

            if isinstance(element, DataSink):
                sinks.add(element)

        LOGGER.info("Beginning construction of type graph")
        # noinspection PyTypeChecker
        self._type_graph = _build_type_graph(sources, sinks, transformers)
        LOGGER.info("Completed construction of type graph")
        self._sources = targets
        self._sinks = sinks
        self._get_types = {}
        self._put_types = {}

    def _transform(self, source_type: Type[S], target_type: Type[T]) -> Tuple[Callable[[S], T], int]:
        try:
            LOGGER.info("Searching type graph for shortest path from \"{source_type}\" to \"{target_type}\"".format(source_type=source_type.__name__, target_type=target_type.__name__))
            path = dijkstra_path(self._type_graph, source=source_type, target=target_type, weight="cost")
            LOGGER.info("Found a path from \"{source_type}\" to \"{target_type}\"".format(source_type=source_type.__name__, target_type=target_type.__name__))
        except (KeyError, NetworkXNoPath):
            raise NoConversionError("Pipeline can't convert \"{source_type}\" to \"{target_type}\"".format(source_type=source_type, target_type=target_type))

        LOGGER.info("Building transformer chain from \"{source_type}\" to \"{target_type}\"".format(source_type=source_type.__name__, target_type=target_type.__name__))
        chain = []
        cost = 0
        for source, target in _pairwise(path):
            transformer = self._type_graph.adj[source][target][_TRANSFORMER]
            chain.append((transformer, target))
            cost += transformer.cost
        LOGGER.info("Built transformer chain from \"{source_type}\" to \"{target_type}\"".format(source_type=source_type.__name__, target_type=target_type.__name__))

        if not chain:
            return _identity, 0

        return partial(_transform, transformer_chain=chain), cost

    def _best_transform_from(self, source_type: Type[S], target_types: Iterable[Type]) -> Tuple[Callable[[S], Any], Type]:
        best = None
        best_cost = _MAX_TRANSFORM_COST
        to_type = None
        for target_type in target_types:
            try:
                transform, cost = self._transform(source_type, target_type)
                if cost < best_cost:
                    best = transform
                    best_cost = cost
                    to_type = target_type
            except NoConversionError:
                pass
        if best is None:
            raise NoConversionError("Pipeline can't convert \"{source_type}\" to any of \"{target_types}\"".format(source_type=source_type, target_types=target_types))
        return best, to_type

    def _best_transform_to(self, target_type: Type[T], source_types: Iterable[Type]) -> Tuple[Callable[[T], Any], Type]:
        best = None
        best_cost = _MAX_TRANSFORM_COST
        from_type = None
        for source_type in source_types:
            try:
                transform, cost = self._transform(source_type, target_type)
                if cost < best_cost:
                    best = transform
                    best_cost = cost
                    from_type = source_type
            except NoConversionError:
                pass
        if best is None:
            raise NoConversionError("Pipeline can't convert from any of \"{source_types}\" to \"{target_type}\"".format(source_types=source_types, target_type=target_type))
        return best, from_type

    def _create_sink_handlers(self, type: Type[T], targets: Iterable[DataSink]) -> Set[DataSink]:
        sink_handlers = set()
        for sink in targets:
            if TYPE_WILDCARD in sink.accepts or type in sink.accepts:
                sink_handlers.add(_SinkHandler(sink, type, _identity))
            else:
                try:
                    transform, store_type = self._best_transform_from(type, sink.accepts)
                    sink_handlers.add(_SinkHandler(sink, store_type, transform))
                except NoConversionError:
                    pass

        return sink_handlers

    def _create_source_handlers(self, type: Type[T]) -> List[_SourceHandler]:
        source_handlers = []
        for source, targets in self._sources:
            if TYPE_WILDCARD in source.provides or type in source.provides:
                sink_handlers = self._create_sink_handlers(type, targets)
                source_handlers.append(_SourceHandler(source, type, _identity, {sink_handler: False for sink_handler in sink_handlers}))
            else:
                try:
                    transform, source_type = self._best_transform_to(type, source.provides)
                    sink_handlers = {sink_handler: False for sink_handler in self._create_sink_handlers(source_type, targets)}
                    sink_handlers.update({sink_handler: True for sink_handler in self._create_sink_handlers(type, targets)})
                    source_handlers.append(_SourceHandler(source, source_type, transform, sink_handlers))
                except NoConversionError:
                    pass

        return source_handlers

    def _get_handlers(self, type: Type[T]) -> List[_SourceHandler]:
        handlers = self._create_source_handlers(type)

        if not handlers:
            raise NoConversionError("No source provides \"{type}\"".format(type=type.__name__))

        return handlers

    def _put_handlers(self, type: Type[T]) -> Set[_SinkHandler]:
        handlers = self._create_sink_handlers(type, self._sinks)

        if not handlers:
            raise NoConversionError("No sink accepts \"{type}\"".format(type=type.__name__))

        return handlers

    def _new_context(self) -> PipelineContext:
        context = PipelineContext()
        context[PipelineContext.Keys.PIPELINE] = self
        return context

    def get(self, type: Type[T], query: Mapping[str, Any]) -> T:
        """Gets a query from the data pipeline.

        1) Extracts the query the sequence of data sources.
        2) Inserts the result into the data sinks (if appropriate).
        3) Transforms the result into the requested type if it wasn't already.
        4) Inserts the transformed result into any data sinks.

        Args:
            query: The query being requested.
            context: The context for the extraction (mutable).

        Returns:
            The requested object.
        """
        LOGGER.info("Getting SourceHandlers for \"{type}\"".format(type=type.__name__))
        try:
            handlers = self._get_types[type]
        except KeyError:
            try:
                LOGGER.info("Building new SourceHandlers for \"{type}\"".format(type=type.__name__))
                handlers = self._get_handlers(type)
            except NoConversionError:
                handlers = None
            self._get_types[type] = handlers

        if handlers is None:
            raise NoConversionError("No source can provide \"{type}\"".format(type=type.__name__))

        LOGGER.info("Creating new PipelineContext")
        context = self._new_context()

        LOGGER.info("Querying SourceHandlers for \"{type}\"".format(type=type.__name__))
        for handler in handlers:
            try:
                return handler.get(query, context)
            except NotFoundError:
                pass

        raise NotFoundError("No source returned a query result!")

    def get_many(self, type: Type[T], query: Mapping[str, Any], streaming: bool = False) -> Iterable[T]:
        """Gets a query from the data pipeline, which contains a request for multiple objects.

        1) Extracts the query the sequence of data sources.
        2) Inserts the results into the data sinks (if appropriate).
        3) Transforms the results into the requested type if it wasn't already.
        4) Inserts the transformed result into any data sinks.

        Args:
            query: The query being requested (contains a request for multiple objects).
            context: The context for the extraction (mutable).
            streaming: Specifies whether the results should be returned as a generator (default False).

        Returns:
            The requested objects or a generator of the objects if streaming is True.
        """
        LOGGER.info("Getting SourceHandlers for \"{type}\"".format(type=type.__name__))
        try:
            handlers = self._get_types[type]
        except KeyError:
            try:
                LOGGER.info("Building new SourceHandlers for \"{type}\"".format(type=type.__name__))
                handlers = self._get_handlers(type)
            except NoConversionError:
                handlers = None
            self._get_types[type] = handlers

        if handlers is None:
            raise NoConversionError("No source can provide \"{type}\"".format(type=type.__name__))

        LOGGER.info("Creating new PipelineContext")
        context = self._new_context()

        LOGGER.info("Querying SourceHandlers for \"{type}\"".format(type=type.__name__))
        for handler in handlers:
            try:
                return handler.get_many(query, context, streaming)
            except NotFoundError:
                pass

        raise NotFoundError("No source returned a query result!")

    def put(self, type: Type[T], item: T) -> None:
        """Puts an objects into the data pipeline. The object may be transformed into a new type for insertion if necessary.

        Args:
            item: The object to be inserted into the data pipeline.
        """
        LOGGER.info("Getting SinkHandlers for \"{type}\"".format(type=type.__name__))
        try:
            handlers = self._put_types[type]
        except KeyError:
            try:
                LOGGER.info("Building new SinkHandlers for \"{type}\"".format(type=type.__name__))
                handlers = self._put_handlers(type)
            except NoConversionError:
                handlers = None
            self._get_types[type] = handlers

        LOGGER.info("Creating new PipelineContext")
        context = self._new_context()

        LOGGER.info("Sending item \"{item}\" to SourceHandlers".format(item=item))
        if handlers is not None:
            for handler in handlers:
                handler.put(item, context)

    def put_many(self, type: Type[T], items: Iterable[T]) -> None:
        """Puts multiple objects of the same type into the data sink. The objects may be transformed into a new type for insertion if necessary.

        Args:
            items: An iterable (e.g. list) of objects to be inserted into the data pipeline.
        """
        LOGGER.info("Getting SinkHandlers for \"{type}\"".format(type=type.__name__))
        try:
            handlers = self._put_types[type]
        except KeyError:
            try:
                LOGGER.info("Building new SinkHandlers for \"{type}\"".format(type=type.__name__))
                handlers = self._put_handlers(type)
            except NoConversionError:
                handlers = None
            self._get_types[type] = handlers

        LOGGER.info("Creating new PipelineContext")
        context = self._new_context()

        LOGGER.info("Sending items \"{items}\" to SourceHandlers".format(items=items))
        if handlers is not None:
            items = list(items)
            for handler in handlers:
                handler.put_many(items, context)



import random
from typing import Type, TypeVar, Iterable

import pytest

from datapipelines import DataSink, CompositeDataSink, PipelineContext

#######################################
# Create simple DataSinks for testing #
#######################################

T = TypeVar("T")

VALUES_COUNT = 100
VALUES_MAX = 100000000


class SimpleWildcardDataSink(DataSink):
    def __init__(self) -> None:
        self.items = {}

    def put(self, type: Type[T], item: T, context: PipelineContext = None) -> None:
        try:
            for_type = self.items[type]
        except KeyError:
            for_type = set()
            self.items[type] = for_type

        for_type.add(item)

    def put_many(self, type: Type[T], items: Iterable[T], context: PipelineContext = None) -> None:
        try:
            for_type = self.items[type]
        except KeyError:
            for_type = set()
            self.items[type] = for_type

        for item in items:
            for_type.add(item)


class IntFloatDataSink(DataSink):
    def __init__(self) -> None:
        self.items = {
            int: set(),
            float: set()
        }

    @DataSink.dispatch
    def put(self, type: Type[T], item: T, context: PipelineContext = None) -> None:
        pass

    @DataSink.dispatch
    def put_many(self, type: Type[T], items: Iterable[T], context: PipelineContext = None) -> None:
        pass

    @put.register(int)
    def put_int(self, item: int, context: PipelineContext = None) -> None:
        self.items[int].add(item)

    @put_many.register(int)
    def put_many_int(self, items: Iterable[int], context: PipelineContext = None) -> None:
        for item in items:
            self.items[int].add(item)

    @put.register(float)
    def put_float(self, item: float, context: PipelineContext = None) -> None:
        self.items[float].add(item)

    @put_many.register(float)
    def put_many_float(self, items: Iterable[float], context: PipelineContext = None) -> None:
        for item in items:
            self.items[float].add(item)


class StringDataSink(DataSink):
    def __init__(self) -> None:
        self.items = {
            str: set()
        }

    @DataSink.dispatch
    def put(self, type: Type[T], item: T, context: PipelineContext = None) -> None:
        pass

    @DataSink.dispatch
    def put_many(self, type: Type[T], items: Iterable[T], context: PipelineContext = None) -> None:
        pass

    @put.register(str)
    def put_int(self, item: str, context: PipelineContext = None) -> None:
        self.items[str].add(item)

    @put_many.register(str)
    def put_many_int(self, items: Iterable[str], context: PipelineContext = None) -> None:
        for item in items:
            self.items[str].add(item)


########################
# Unsupported Function #
########################

def test_unsupported():
    from datapipelines import UnsupportedError

    unsupported = DataSink.unsupported(int)
    assert type(unsupported) is UnsupportedError

    unsupported = DataSink.unsupported(float)
    assert type(unsupported) is UnsupportedError

    unsupported = DataSink.unsupported(str)
    assert type(unsupported) is UnsupportedError


####################
# Accepts Function #
####################

def test_accepts():
    sink = IntFloatDataSink()

    assert sink.accepts == {int, float}


def test_wildcard_accepts():
    from datapipelines import TYPE_WILDCARD
    sink = SimpleWildcardDataSink()

    assert sink.accepts is TYPE_WILDCARD


################
# Put Function #
################

def test_put():
    sink = IntFloatDataSink()

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        result = sink.put(int, value)

        assert result is None
        assert value in sink.items[int]

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        result = sink.put(float, value)

        assert result is None
        assert value in sink.items[float]


def test_put_unsupported():
    from datapipelines import UnsupportedError
    sink = IntFloatDataSink()

    with pytest.raises(UnsupportedError):
        sink.put(str, "test")


def test_wildcard_put():
    sink = SimpleWildcardDataSink()

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        result = sink.put(int, value)

        assert result is None
        assert value in sink.items[int]

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        result = sink.put(float, value)

        assert result is None
        assert value in sink.items[float]


#####################
# Put Many Function #
#####################

def test_put_many():
    sink = IntFloatDataSink()

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        items = (value for _ in range(VALUES_COUNT))
        result = sink.put_many(int, items)

        assert result is None
        assert value in sink.items[int]

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        items = (value for _ in range(VALUES_COUNT))
        result = sink.put_many(float, items)

        assert result is None
        assert value in sink.items[float]


def test_put_many_unsupported():
    from datapipelines import UnsupportedError
    sink = IntFloatDataSink()

    with pytest.raises(UnsupportedError):
        sink.put_many(str, ("test" for _ in range(VALUES_COUNT)))


def test_wildcard_put_many():
    sink = SimpleWildcardDataSink()

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        items = (value for _ in range(VALUES_COUNT))
        result = sink.put_many(int, items)

        assert result is None
        assert value in sink.items[int]

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        items = (value for _ in range(VALUES_COUNT))
        result = sink.put_many(float, items)

        assert result is None
        assert value in sink.items[float]


#####################
# CompositeDataSink #
#####################

def test_composite_accepts():
    int_float = IntFloatDataSink()
    string = StringDataSink()
    sink = CompositeDataSink({int_float, string})
    assert sink.accepts == {int, float, str}


def test_composite_put():
    int_float = IntFloatDataSink()
    string = StringDataSink()
    sink = CompositeDataSink({int_float, string})

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        result = sink.put(int, value)

        assert result is None
        assert value in int_float.items[int]

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        result = sink.put(float, value)

        assert result is None
        assert value in int_float.items[float]

    values = [str(random.uniform(-VALUES_MAX, VALUES_MAX)) for _ in range(VALUES_COUNT)]

    for value in values:
        result = sink.put(str, value)

        assert result is None
        assert value in string.items[str]


def test_composite_put_unsupported():
    from datapipelines import UnsupportedError
    int_float = IntFloatDataSink()
    string = StringDataSink()
    sink = CompositeDataSink({int_float, string})

    with pytest.raises(UnsupportedError):
        sink.put(bytes, bytes())


def test_composite_put_many():
    int_float = IntFloatDataSink()
    string = StringDataSink()
    sink = CompositeDataSink({int_float, string})

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        items = (value for _ in range(VALUES_COUNT))
        result = sink.put_many(int, items)

        assert result is None
        assert value in int_float.items[int]

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        items = (value for _ in range(VALUES_COUNT))
        result = sink.put_many(float, items)

        assert result is None
        assert value in int_float.items[float]

    values = [str(random.uniform(-VALUES_MAX, VALUES_MAX)) for _ in range(VALUES_COUNT)]

    for value in values:
        items = (value for _ in range(VALUES_COUNT))
        result = sink.put_many(str, items)

        assert result is None
        assert value in string.items[str]


def test_composite_put_many_unsupported():
    from datapipelines import UnsupportedError
    int_float = IntFloatDataSink()
    string = StringDataSink()
    sink = CompositeDataSink({int_float, string})

    with pytest.raises(UnsupportedError):
        sink.put_many(bytes, (bytes() for _ in range(VALUES_COUNT)))



from typing import Generic, TypeVar, Any

TYPE_WILDCARD = Any


class UnsupportedError(ValueError):
    pass


class NotFoundError(ValueError):
    pass


class PipelineContext(dict):
    class Keys(object):
        PIPELINE = "pipeline"
        EXPIRATION = "expires"


T = TypeVar("T")
Q = TypeVar("Q")


class TypePair(Generic[T, Q]):
    pass



import random
from typing import Type, TypeVar, Mapping, Any, Iterable, Generator

import pytest
from networkx import DiGraph

from datapipelines import DataPipeline, DataSource, DataSink, DataTransformer, PipelineContext, NotFoundError, NoConversionError
from datapipelines.pipelines import _build_type_graph, _pairwise, _identity, _transform, _SinkHandler, _SourceHandler

T = TypeVar("T")
F = TypeVar("F")

VALUE_KEY = "value"
COUNT_KEY = "count"

VALUES_COUNT = 100
VALUES_MAX = 100000000

# Seriously where is this in the std lib...
GENERATOR_CLASS = (None for _ in range(0)).__class__


###############################################
# Create simple pipeline elements for testing #
###############################################

class IntSource(DataSource):
    @DataSource.dispatch
    def get(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    @get.register(int)
    def get_int(self, query: Mapping[str, Any], context: PipelineContext = None) -> int:
        value = query.get(VALUE_KEY)

        try:
            return int(value)
        except ValueError:
            raise NotFoundError("Couldn't cast the query value to \"int\"")

    @get_many.register(int)
    def get_many_int(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[int, None, None]:
        value = query.get(VALUE_KEY)
        count = query.get(COUNT_KEY)

        try:
            value = int(value)
        except ValueError:
            raise NotFoundError("Couldn't cast the query value to \"int\"")

        return (value for _ in range(count))


class FloatStore(DataSource, DataSink):
    def __init__(self) -> None:
        self.items = set()

    @DataSink.dispatch
    def put(self, type: Type[T], item: T, context: PipelineContext = None) -> None:
        pass

    @DataSink.dispatch
    def put_many(self, type: Type[T], items: Iterable[T], context: PipelineContext = None) -> None:
        pass

    @put.register(float)
    def put_float(self, item: float, context: PipelineContext = None) -> None:
        self.items.add(item)

    @put_many.register(float)
    def put_many_float(self, items: Iterable[float], context: PipelineContext = None) -> None:
        for item in items:
            self.items.add(item)

    @DataSource.dispatch
    def get(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSink.dispatch
    def get_many(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    @get.register(float)
    def get_float(self, query: Mapping[str, Any], context: PipelineContext = None) -> float:
        value = query.get(VALUE_KEY)

        try:
            value = float(value)
        except ValueError:
            raise NotFoundError("Couldn't cast the query value to \"float\"")

        if value not in self.items:
            raise NotFoundError("Query value wasn't in store!")

        return value

    @get_many.register(float)
    def get_many_float(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[float, None, None]:
        value = query.get(VALUE_KEY)
        count = query.get(COUNT_KEY)

        try:
            value = float(value)
        except ValueError:
            raise NotFoundError("Couldn't cast the query value to \"float\"")

        if value not in self.items:
            raise NotFoundError("Query value wasn't in store!")

        return (value for _ in range(count))


class IntFloatTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    @transform.register(int, float)
    def int_to_float(self, value: int, context: PipelineContext = None) -> float:
        return float(value)


class FloatIntTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    @transform.register(float, int)
    def float_to_int(self, value: float, context: PipelineContext = None) -> int:
        return int(value)

    @property
    def cost(self) -> int:
        return 3


class StringTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    @transform.register(int, str)
    def int_to_str(self, value: int, context: PipelineContext = None) -> str:
        return str(value)

    @transform.register(float, str)
    def float_to_str(self, value: float, context: PipelineContext = None) -> str:
        return str(value)

    @transform.register(str, int)
    def str_to_int(self, value: str, context: PipelineContext = None) -> int:
        return int(float(value))

    @transform.register(str, float)
    def str_to_float(self, value: str, context: PipelineContext = None) -> float:
        return float(value)


#########################
# Type graph generation #
#########################

def test_build_type_graph():
    int_source = IntSource()
    float_store = FloatStore()
    int_float = IntFloatTransformer()
    float_int = FloatIntTransformer()
    string = StringTransformer()

    sources = {int_source, float_store}
    sinks = {float_store}
    transformers = {int_float, float_int, string}

    expected = DiGraph()
    expected.add_node(str)
    expected.add_node(int, sources={int_source})
    expected.add_node(float, sources={float_store}, sinks={float_store})
    expected.add_edge(int, float, cost=1, transformer=int_float)
    expected.add_edge(float, int, cost=3, transformer=float_int)
    expected.add_edge(str, int, cost=1, transformer=string)
    expected.add_edge(str, float, cost=1, transformer=string)
    expected.add_edge(int, str, cost=1, transformer=string)
    expected.add_edge(float, str, cost=1, transformer=string)

    # noinspection PyTypeChecker
    actual = _build_type_graph(sources, sinks, transformers)

    for node in expected.nodes():
        assert expected.node[node] == actual.node[node]

    for source, target in expected.edges():
        assert expected.edge[source][target] == actual.edge[source][target]

    for node in actual.nodes():
        assert expected.node[node] == actual.node[node]

    for source, target in actual.edges():
        assert expected.edge[source][target] == actual.edge[source][target]


#####################
# Utility functions #
#####################

def test_pairwise():
    values = (x for x in range(VALUES_COUNT))
    values = _pairwise(values)

    for i in range(VALUES_COUNT - 1):
        assert next(values) == (i, i + 1)


def test_identity():
    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        assert _identity(value) == value

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        assert _identity(value) == value


def test_transform():
    string = StringTransformer()
    int_float = IntFloatTransformer()
    float_int = FloatIntTransformer()

    chain = [(string, str), (string, float), (float_int, int), (string, str), (string, float), (float_int, int), (int_float, float)]
    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]
    for value in values:
        result = _transform(chain, value)

        assert type(result) is float
        assert result == value


###############
# SinkHandler #
###############

def test_sink_handler_put():
    sink = FloatStore()

    def convert(data: int, context: PipelineContext = None) -> float:
        return float(data)

    handler = _SinkHandler(sink, float, convert)

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        result = handler.put(value)

        assert result is None
        assert value in sink.items


def test_sink_handler_put_many():
    sink = FloatStore()

    def convert(data: int, context: PipelineContext = None) -> float:
        return float(data)

    handler = _SinkHandler(sink, float, convert)

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        items = (value for _ in range(VALUES_COUNT))
        result = handler.put_many(items)

        assert result is None
        assert value in sink.items


#################
# SourceHandler #
#################

def test_source_handler_get():
    source = IntSource()
    before_sink = FloatStore()
    after_sink = FloatStore()

    def convert(data: int, context: PipelineContext = None) -> float:
        return float(data)

    sinks = {
        _SinkHandler(before_sink, float, convert): False,
        _SinkHandler(after_sink, float, _identity): True
    }

    handler = _SourceHandler(source, int, convert, sinks)

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value}
        result = handler.get(query)

        assert type(result) is float
        floored_value = float(int(value))  # float to int conversion will floor the value
        assert result == floored_value

        assert floored_value in before_sink.items
        assert floored_value in after_sink.items


def test_source_handler_get_many_non_streaming():
    source = IntSource()
    before_sink = FloatStore()
    after_sink = FloatStore()

    def convert(data: int, context: PipelineContext = None) -> float:
        return float(data)

    sinks = {
        _SinkHandler(before_sink, float, convert): False,
        _SinkHandler(after_sink, float, _identity): True
    }

    handler = _SourceHandler(source, int, convert, sinks)

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value, COUNT_KEY: VALUES_COUNT}
        result = handler.get_many(query, streaming=False)

        assert type(result) is list
        floored_value = float(int(value))  # float to int conversion will floor the value
        for res in result:
            assert res == floored_value

            assert floored_value in before_sink.items
            assert floored_value in after_sink.items


def test_source_handler_get_many_streaming():
    source = IntSource()
    before_sink = FloatStore()
    after_sink = FloatStore()

    def convert(data: int, context: PipelineContext = None) -> float:
        return float(data)

    sinks = {
        _SinkHandler(before_sink, float, convert): False,
        _SinkHandler(after_sink, float, _identity): True
    }

    handler = _SourceHandler(source, int, convert, sinks)

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value, COUNT_KEY: VALUES_COUNT}
        result = handler.get_many(query, streaming=True)

        assert type(result) is GENERATOR_CLASS
        floored_value = float(int(value))  # float to int conversion will floor the value
        for res in result:
            assert res == floored_value

            assert floored_value in before_sink.items
            assert floored_value in after_sink.items


################
# DataPipeline #
################

def test_pipeline_transform():
    int_source = IntSource()
    float_store = FloatStore()
    int_float = IntFloatTransformer()
    float_int = FloatIntTransformer()
    string = StringTransformer()

    elements = [float_store, int_source]
    transformers = {int_float, float_int, string}

    # noinspection PyTypeChecker
    pipeline = DataPipeline(elements, transformers)

    # Identity of type in graph
    transform, cost = pipeline._transform(int, int)
    assert cost == 0
    assert transform is _identity

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        result = transform(data=value)

        assert type(result) is int
        assert result == value

    # Identity of type not in graph
    transform, cost = pipeline._transform(bool, bool)
    assert cost == 0
    assert transform is _identity

    values = [random.choice([True, False]) for _ in range(VALUES_COUNT)]

    for value in values:
        result = transform(data=value)

        assert type(result) is bool
        assert result == value

    # Simple transform
    transform, cost = pipeline._transform(int, float)
    assert cost == 1

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        result = transform(data=value)

        assert type(result) is float
        assert result == value

    # Avoid expensive transformer
    transform, cost = pipeline._transform(float, int)
    assert cost == 2  # It will go through the StringTransformer twice, which is cheaper than the FloatIntTransformer

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        result = transform(data=value)

        assert type(result) is int
        assert result == int(value)


def test_pipeline_transform_impossible():
    int_source = IntSource()
    float_store = FloatStore()
    int_float = IntFloatTransformer()

    elements = [float_store, int_source]
    transformers = {int_float}

    # noinspection PyTypeChecker
    pipeline = DataPipeline(elements, transformers)

    with pytest.raises(NoConversionError):
        pipeline._transform(float, int)

    with pytest.raises(NoConversionError):
        pipeline._transform(str, int)

    with pytest.raises(NoConversionError):
        pipeline._transform(int, str)


def test_get_handlers():
    int_source = IntSource()
    float_store = FloatStore()
    int_float = IntFloatTransformer()
    float_int = FloatIntTransformer()
    string = StringTransformer()

    elements = [float_store, int_source]
    transformers = {int_float, float_int, string}

    # noinspection PyTypeChecker
    pipeline = DataPipeline(elements, transformers)

    handlers = pipeline._get_handlers(str)
    assert type(handlers) is list
    assert len(handlers) is 2

    assert handlers[0]._source is float_store
    assert handlers[0]._source_type is float

    assert handlers[1]._source is int_source
    assert handlers[1]._source_type is int

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]
    for value in values:
        assert type(handlers[0]._transform(data=value)) is str

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]
    for value in values:
        assert type(handlers[1]._transform(data=value)) is str

    assert len(handlers[0]._before_transform) == 0
    assert len(handlers[0]._after_transform) == 0

    assert len(handlers[1]._before_transform) == 1
    assert len(handlers[1]._after_transform) == 1

    sink_handler = list(handlers[1]._before_transform)[0]

    assert sink_handler._sink is float_store
    assert sink_handler._store_type is float

    for value in values:
        assert type(sink_handler._transform(data=value)) is float


def test_get_handlers_impossible():
    int_source = IntSource()
    float_store = FloatStore()
    int_float = IntFloatTransformer()

    elements = [float_store, int_source]
    transformers = {int_float}

    # noinspection PyTypeChecker
    pipeline = DataPipeline(elements, transformers)

    with pytest.raises(NoConversionError):
        pipeline._get_handlers(str)


def test_put_handlers():
    int_source = IntSource()
    float_store = FloatStore()
    int_float = IntFloatTransformer()
    float_int = FloatIntTransformer()
    string = StringTransformer()

    elements = [float_store, int_source]
    transformers = {int_float, float_int, string}

    # noinspection PyTypeChecker
    pipeline = DataPipeline(elements, transformers)

    handlers = pipeline._put_handlers(str)
    assert type(handlers) is set
    assert len(handlers) is 1

    handler = list(handlers)[0]

    assert handler._sink is float_store
    assert handler._store_type is float

    values = [str(random.uniform(-VALUES_MAX, VALUES_MAX)) for _ in range(VALUES_COUNT)]

    for value in values:
        assert type(handler._transform(data=value)) is float


def test_put_handlers_impossible():
    int_source = IntSource()
    float_store = FloatStore()
    int_float = IntFloatTransformer()

    elements = [float_store, int_source]
    transformers = {int_float}

    # noinspection PyTypeChecker
    pipeline = DataPipeline(elements, transformers)

    with pytest.raises(NoConversionError):
        pipeline._put_handlers(str)


def test_new_context():
    int_source = IntSource()
    float_store = FloatStore()
    int_float = IntFloatTransformer()
    float_int = FloatIntTransformer()
    string = StringTransformer()

    elements = [float_store, int_source]
    transformers = {int_float, float_int, string}

    # noinspection PyTypeChecker
    pipeline = DataPipeline(elements, transformers)

    context = pipeline._new_context()
    assert type(context) is PipelineContext
    assert context[PipelineContext.Keys.PIPELINE] is pipeline


def test_get():
    int_source = IntSource()
    float_store = FloatStore()
    int_float = IntFloatTransformer()
    float_int = FloatIntTransformer()
    string = StringTransformer()

    elements = [float_store, int_source]
    transformers = {int_float, float_int, string}

    # noinspection PyTypeChecker
    pipeline = DataPipeline(elements, transformers)

    values = [str(random.randint(-VALUES_MAX, VALUES_MAX)) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value}

        result = pipeline.get(int, query)
        assert type(result) is int
        assert result == int(value)

        assert float(value) in float_store.items

    float_store.items.clear()

    for value in values:
        query = {VALUE_KEY: value}

        result = pipeline.get(float, query)
        assert type(result) is float
        assert result == float(value)

        assert float(value) in float_store.items

    float_store.items.clear()

    for value in values:
        query = {VALUE_KEY: value}

        result = pipeline.get(str, query)
        assert type(result) is str
        assert result == value

        assert float(value) in float_store.items


def test_get_many_non_streaming():
    int_source = IntSource()
    float_store = FloatStore()
    int_float = IntFloatTransformer()
    float_int = FloatIntTransformer()
    string = StringTransformer()

    elements = [float_store, int_source]
    transformers = {int_float, float_int, string}

    # noinspection PyTypeChecker
    pipeline = DataPipeline(elements, transformers)

    values = [str(random.randint(-VALUES_MAX, VALUES_MAX)) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value, COUNT_KEY: VALUES_COUNT}

        result = pipeline.get_many(int, query, streaming=False)
        assert type(result) is list

        for res in result:
            assert type(res) is int
            assert res == int(value)
            assert float(value) in float_store.items

    float_store.items.clear()

    for value in values:
        query = {VALUE_KEY: value, COUNT_KEY: VALUES_COUNT}

        result = pipeline.get_many(float, query, streaming=False)
        assert type(result) is list

        for res in result:
            assert type(res) is float
            assert res == float(value)
            assert float(value) in float_store.items

    float_store.items.clear()

    for value in values:
        query = {VALUE_KEY: value, COUNT_KEY: VALUES_COUNT}

        result = pipeline.get_many(str, query, streaming=False)
        assert type(result) is list

        for res in result:
            assert type(res) is str
            assert res == value
            assert float(value) in float_store.items


def test_get_many_streaming():
    int_source = IntSource()
    float_store = FloatStore()
    int_float = IntFloatTransformer()
    float_int = FloatIntTransformer()
    string = StringTransformer()

    elements = [float_store, int_source]
    transformers = {int_float, float_int, string}

    # noinspection PyTypeChecker
    pipeline = DataPipeline(elements, transformers)

    values = [str(random.randint(-VALUES_MAX, VALUES_MAX)) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value, COUNT_KEY: VALUES_COUNT}

        result = pipeline.get_many(int, query, streaming=True)
        assert type(result) is GENERATOR_CLASS

        for res in result:
            assert type(res) is int
            assert res == int(value)
            assert float(value) in float_store.items

    float_store.items.clear()

    for value in values:
        query = {VALUE_KEY: value, COUNT_KEY: VALUES_COUNT}

        result = pipeline.get_many(float, query, streaming=True)
        assert type(result) is GENERATOR_CLASS

        for res in result:
            assert type(res) is float
            assert res == float(value)
            assert float(value) in float_store.items

    float_store.items.clear()

    for value in values:
        query = {VALUE_KEY: value, COUNT_KEY: VALUES_COUNT}

        result = pipeline.get_many(str, query, streaming=True)
        assert type(result) is GENERATOR_CLASS

        for res in result:
            assert type(res) is str
            assert res == value
            assert float(value) in float_store.items


def test_put():
    int_source = IntSource()
    float_store = FloatStore()
    int_float = IntFloatTransformer()
    float_int = FloatIntTransformer()
    string = StringTransformer()

    elements = [float_store, int_source]
    transformers = {int_float, float_int, string}

    # noinspection PyTypeChecker
    pipeline = DataPipeline(elements, transformers)

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        result = pipeline.put(int, value)

        assert result is None
        assert float(value) in float_store.items

    float_store.items.clear()

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        result = pipeline.put(float, value)

        assert result is None
        assert value in float_store.items

    float_store.items.clear()

    values = [str(random.uniform(-VALUES_MAX, VALUES_MAX)) for _ in range(VALUES_COUNT)]

    for value in values:
        result = pipeline.put(str, value)

        assert result is None
        assert float(value) in float_store.items


def test_put_many():
    int_source = IntSource()
    float_store = FloatStore()
    int_float = IntFloatTransformer()
    float_int = FloatIntTransformer()
    string = StringTransformer()

    elements = [float_store, int_source]
    transformers = {int_float, float_int, string}

    # noinspection PyTypeChecker
    pipeline = DataPipeline(elements, transformers)

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        items = (value for _ in range(VALUES_COUNT))
        result = pipeline.put_many(int, items)

        assert result is None
        assert float(value) in float_store.items

    float_store.items.clear()

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        items = (value for _ in range(VALUES_COUNT))
        result = pipeline.put_many(float, items)

        assert result is None
        assert value in float_store.items

    float_store.items.clear()

    values = [str(random.uniform(-VALUES_MAX, VALUES_MAX)) for _ in range(VALUES_COUNT)]

    for value in values:
        items = (value for _ in range(VALUES_COUNT))
        result = pipeline.put_many(str, items)

        assert result is None
        assert float(value) in float_store.items



from .common import PipelineContext, UnsupportedError, NotFoundError, TYPE_WILDCARD
from .pipelines import DataPipeline, NoConversionError
from .queries import Query, QueryValidationError, QueryValidatorStructureError, validate_query
from .sinks import DataSink, CompositeDataSink
from .sources import DataSource, CompositeDataSource
from .transformers import DataTransformer, CompositeDataTransformer

__all__ = ["DataTransformer", "CompositeDataTransformer", "DataPipeline", "NoConversionError", "Query", "QueryValidationError", "QueryValidatorStructureError", "validate_query", "DataSource", "CompositeDataSource", "DataSink", "CompositeDataSink", "PipelineContext", "UnsupportedError", "NotFoundError", "TYPE_WILDCARD"]



from abc import ABC, abstractmethod
from enum import Enum
from copy import deepcopy
from typing import Type, MutableMapping, Any, Iterable, Union, Callable
from functools import wraps

from .pipelines import PipelineContext


class QueryValidationError(ValueError):
    pass


class MissingKeyError(QueryValidationError):
    pass


class WrongValueTypeError(QueryValidationError):
    pass


class BoundKeyExistenceError(QueryValidationError):
    pass


class QueryValidatorStructureError(AttributeError):
    pass


class _ValidationNode(ABC):
    @abstractmethod
    def evaluate(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> bool:
        pass

    @property
    @abstractmethod
    def falsifiable(self):
        pass


class _RootNode(_ValidationNode):
    def __init__(self, children: Iterable[Union["_KeyNode", "_AndNode", "_OrNode"]] = None) -> None:
        if children is None:
            children = list()
        self.children = list(children)

    def __str__(self) -> str:
        return " ALSO ".join(str(child) for child in self.children)

    @property
    def falsifiable(self):
        return False

    def evaluate(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> True:
        # This always returns true because it'll raise a QueryValidationError if the query wasn't
        # valid. We also don't want normal AND behavior at this level so we just evaluate the
        # children and return True
        for child in self.children:
            child.evaluate(query, context)
        return True


class _AndNode(_ValidationNode):
    def __init__(self, children: Iterable[Union["_KeyNode", "_AndNode", "_OrNode"]]) -> None:
        self.children = list(children)

    def __str__(self) -> str:
        return " AND ".join(str(child) for child in self.children)

    @property
    def falsifiable(self):
        return False

    def evaluate(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> True:
        # This is a little weird. We have to handle the case of can_have("x").and_("y") here.
        # Since the only type of Node that can return False rather than just raising a
        # QueryValidationError is a KeyNode that isn't required, or an OrNode whose children
        # are all KeyNodes that aren't required, we can't short circuit on a False value. If
        # we have can_have("x").and_("y"), and only some are True, we want raise a
        # QueryValidationError. If all are True or all are False, everything's fine.
        all_true = True
        all_possible_false = True
        for child in self.children:
            is_true = child.evaluate(query, context)
            all_true = all_true and is_true
            if child.falsifiable:
                all_possible_false = all_possible_false and not is_true

        if all_possible_false or all_true:
            return True
        raise BoundKeyExistenceError("Query must have all or none of the elements joined with \"and\" in a \"can_have\" statement!")


class _OrNode(_ValidationNode):
    def __init__(self, children: Iterable[Union["_KeyNode", "_AndNode", "_OrNode"]]) -> None:
        self.children = list(children)

    def __str__(self) -> str:
        return " OR ".join(str(child) for child in self.children)

    @property
    def falsifiable(self):
        return True

    def evaluate(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> bool:
        # We can't short circuit this because we want to raise any WrongValueTypeError or BoundKeyExistenceError that could occur.
        # We count failures because the only other node than can return False is a KeyNode that isn't required. An OrNode is only
        # allowed to return False when all its children are False KeyNodes. The AndNodes handle raising an error if a can_have
        # expression with ands and ors fails.
        errors = []
        failure_count = 0
        result = False
        for child in self.children:
            try:
                is_true = child.evaluate(query, context)
                if not is_true:
                    failure_count += 1
                result = is_true or result
            except (MissingKeyError, BoundKeyExistenceError) as error:
                errors.append(error)
                failure_count += 1
        if failure_count == len(self.children) and len(errors) > 0:
            raise errors[0]
        return result


class _DefaultValueNode(_ValidationNode):
    def __init__(self, key: str, value: Union[Any, Callable[[MutableMapping[str, Any]], Any]], supplies_type: Type = None) -> None:
        self.key = key
        self.value = value
        self.supplies_type = supplies_type

    def __str__(self) -> str:
        return self.value

    @property
    def falsifiable(self):
        return False

    def evaluate(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> True:
        if self.supplies_type:
            query[self.key] = self.value(query, context)
        else:
            query[self.key] = deepcopy(self.value)
        return True


class _TypeNode(_ValidationNode):
    def __init__(self, key: str, types: Iterable[Type], child: _DefaultValueNode = None) -> None:
        self.key = key
        self.child = child
        self.types = types

    def __str__(self) -> str:
        return " OR ".join(type.__name__ for type in self.types)

    @property
    def falsifiable(self):
        return False

    def evaluate(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> True:
        try:
            value = query[self.key]
            for type in self.types:
                if issubclass(type, Enum) and isinstance(value, str):
                    value = type(value)
                if isinstance(value, type):
                    query[self.key] = value
                    return True
            raise WrongValueTypeError("{key} must be of type {type} in query!".format(key=self.key, type=self))
        except KeyError:
            if self.child:
                self.child.evaluate(query, context)
        return True


class _KeyNode(_ValidationNode):
    def __init__(self, key: str, required: bool = True, child: _TypeNode = None) -> None:
        self.key = key
        self.required = required
        self.child = child

    def __str__(self) -> str:
        return self.key

    @property
    def falsifiable(self):
        return not self.required

    def evaluate(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> bool:
        has_key = self.key in query
        if self.required and not has_key:
            raise MissingKeyError("{key} must be in query!".format(key=self.key))
        if self.child:
            self.child.evaluate(query, context)
        return has_key


class QueryValidator(object):
    def __init__(self) -> None:
        self._root = _RootNode()
        self._current = None  # type: _KeyNode
        self._parent = None  # type: Union[_AndNode, _OrNode]

    def __call__(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> True:
        return self._root.evaluate(query, context)

    def has(self, key: str) -> "QueryValidator":
        if self._current is not None:
            raise QueryValidatorStructureError("A key is already selected! Try using \"also\" before \"has\".")

        has_node = _KeyNode(key, True)
        self._root.children.append(has_node)
        self._current = has_node
        self._parent = self._root
        return self

    def can_have(self, key: str) -> "QueryValidator":
        if self._current is not None:
            raise QueryValidatorStructureError("A key is already selected! Try using \"also\" before \"can_have\".")

        has_node = _KeyNode(key, False)
        self._root.children.append(has_node)
        self._current = has_node
        self._parent = self._root
        return self

    def as_(self, type: Type) -> "QueryValidator":
        if self._current is None or self._current.child is not None:
            raise QueryValidatorStructureError("No key is selected! Try using \"has\" or \"can_have\" before \"as\".")

        type_node = _TypeNode(self._current.key, {type})
        self._current.child = type_node
        return self

    def as_any_of(self, types: Iterable[Type]) -> "QueryValidator":
        if self._current is None or self._current.child is not None:
            raise QueryValidatorStructureError("No key is selected! Try using \"has\" or \"can_have\" before \"as_any_of\".")

        type_node = _TypeNode(self._current.key, types)
        self._current.child = type_node
        return self

    def or_(self, key: str) -> "QueryValidator":
        if self._current is None:
            raise QueryValidatorStructureError("No key is selected! Try using \"has\" or \"can_have\" before \"or\".")

        if isinstance(self._parent, _OrNode):
            or_node = self._parent
        else:
            or_node = _OrNode({self._current})
            self._parent.children.remove(self._current)
            self._parent.children.append(or_node)
            self._parent = or_node

        has_node = _KeyNode(key, self._current.required)
        or_node.children.append(has_node)
        self._current = has_node
        return self

    def and_(self, key: str) -> "QueryValidator":
        if self._current is None:
            raise QueryValidatorStructureError("No key is selected! Try using \"has\" or \"can_have\" before \"and\".")

        if isinstance(self._parent, _AndNode):
            and_node = self._parent
        else:
            and_node = _AndNode({self._current})
            self._parent.children.remove(self._current)
            self._parent.children.append(and_node)
            self._parent = and_node

        has_node = _KeyNode(key, self._current.required)
        and_node.children.append(has_node)
        self._current = has_node
        return self

    @property
    def also(self) -> "QueryValidator":
        if self._current is None:
            raise QueryValidatorStructureError("No key is selected! Try using \"has\" or \"can_have\" before \"also\".")

        self._current = None
        self._parent = None
        return self

    def with_default(self, value: Union[Any, Callable[[MutableMapping[str, Any]], Any]], supplies_type: Type = None) -> "QueryValidator":
        if self._current is None or self._current.child is not None:
            raise QueryValidatorStructureError("No key is selected! Try using \"can_have\" before \"with_default\".")

        if self._current.required:
            raise QueryValidatorStructureError("Can't assign a default value to a required key! Try using \"can_have\" instead of \"have\".")

        if supplies_type:
            expected_type = supplies_type
        else:
            expected_type = type(value)

        default_node = _DefaultValueNode(self._current.key, value, supplies_type)
        result = self.as_(expected_type)
        result._current.child.child = default_node
        return result


class Query(dict):
    @staticmethod
    def has(key: str) -> QueryValidator:
        return QueryValidator().has(key)

    @staticmethod
    def can_have(key: str) -> QueryValidator:
        return QueryValidator().can_have(key)


def validate_query(validator: QueryValidator, *pre_transforms: Callable[[MutableMapping], None]) -> Callable[[Callable[[Any, MutableMapping[str, Any], PipelineContext], Union[Any, Iterable[Any]]]], Callable[[Any, MutableMapping[str, Any], PipelineContext], Union[Any, Iterable[Any]]]]:
    def wrapper(method: Callable[[Any, MutableMapping[str, Any], PipelineContext], Union[Any, Iterable[Any]]]) -> Callable[[Any, MutableMapping[str, Any], PipelineContext], Union[Any, Iterable[Any]]]:
        @wraps(method)
        def wrapped(self: Any, query: MutableMapping[str, Any], context: PipelineContext = None):
            for transform in pre_transforms:
                transform(query)

            validator(query)
            return method(self, query, context)

        return wrapped
    return wrapper



from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.rst"), encoding="UTF-8") as readme:
    long_description = readme.read()

setup(
    name="datapipelines",
    version="1.0.3",
    author="Meraki Analytics Team",
    author_email="team@merakianalytics.com",
    url="https://github.com/meraki-analytics/datapipelines",
    description="Caching abstraction layer for orchestrating multiple cache tiers",
    long_description=long_description,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Database :: Front-Ends"
    ],
    license="MIT",
    packages=find_packages(),
    zip_safe=True,
    install_requires=[
        "merakicommons", "networkx"
    ],
    extras_require={
        "testing": ["pytest", "flake8"]
    }
)



from abc import ABC, abstractmethod
from functools import singledispatch, update_wrapper
from typing import TypeVar, Type, Callable, Any, Mapping, Iterable, Dict, Set

from merakicommons.cache import lazy_property

from .common import PipelineContext, UnsupportedError, TypePair

F = TypeVar("F")
T = TypeVar("T")


class DataTransformer(ABC):
    @staticmethod
    def unsupported(target_type: Type[T], value: F) -> UnsupportedError:
        return UnsupportedError("The conversion from type \"{from_type}\" to type \"{to_type}\" is not supported by this DataTransformer!".format(from_type=value.__class__.__name__, to_type=target_type.__name__))

    @property
    def transforms(self) -> Mapping[Type, Iterable[Type]]:
        """The available data transformers."""
        try:
            return getattr(self.__class__, "transform")._transforms
        except AttributeError:
            return {}

    @abstractmethod
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        """Transforms an object to a new type.

        Args:
            target_type: The type to be converted to.
            value: The object to be transformed.
            context: The context of the transformation (mutable).
        """
        pass

    @property
    def cost(self) -> int:
        """The cost of the tranformation (default 1)."""
        return 1

    @staticmethod
    def dispatch(method: Callable[[Any, Type[T], F, PipelineContext], T]) -> Callable[[Any, Type[T], F, PipelineContext], T]:
        dispatcher = singledispatch(method)
        transforms = {}

        def wrapper(self: Any, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
            call = dispatcher.dispatch(TypePair[value.__class__, target_type])
            try:
                return call(self, value, context=context)
            except TypeError:
                raise DataTransformer.unsupported(target_type, value)

        def register(from_type: Type[F], to_type: Type[T]) -> Callable[[Any, Type[T], F, PipelineContext], T]:
            try:
                target_types = transforms[from_type]
            except KeyError:
                target_types = set()
                transforms[from_type] = target_types
            target_types.add(to_type)

            return dispatcher.register(TypePair[from_type, to_type])

        wrapper.register = register
        wrapper._transforms = transforms
        update_wrapper(wrapper, method)
        return wrapper


class CompositeDataTransformer(DataTransformer):
    def __init__(self, transformers: Iterable[DataTransformer]) -> None:
        self._transformers = {}
        for transformer in transformers:
            for source, targets in transformer.transforms.items():
                for target in targets:
                    try:
                        current_transformer = self._transformers[(source, target)]
                        if transformer.cost < current_transformer.cost:
                            self._transformers[(source, target)] = transformer
                    except KeyError:
                        self._transformers[(source, target)] = transformer

    @lazy_property
    def transforms(self) -> Dict[Type, Set[Type]]:
        transforms = {}
        for source, target in self._transformers:
            try:
                current = transforms[source]
            except KeyError:
                current = set()
                transforms[source] = current
            current.add(target)
        return transforms

    @lazy_property
    def cost(self) -> int:
        return max(transformer.cost for transformer in self._transformers.values())

    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        try:
            transformer = self._transformers[type(value), target_type]
        except KeyError as error:
            raise DataTransformer.unsupported(target_type, value) from error

        return transformer.transform(target_type, value, context)



import random
from typing import Type, TypeVar, Mapping, Any, Iterable, Generator

import pytest

from datapipelines import DataSource, CompositeDataSource, PipelineContext, NotFoundError

#########################################
# Create simple DataSources for testing #
#########################################

T = TypeVar("T")

VALUE_KEY = "value"
COUNT_KEY = "count"

VALUES_COUNT = 100
VALUES_MAX = 100000000

# Seriously where is this in the std lib...
GENERATOR_CLASS = (None for _ in range(0)).__class__


class SimpleWildcardDataSource(DataSource):
    def get(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> T:
        value = query.get(VALUE_KEY)

        try:
            # noinspection PyCallingNonCallable
            return type(value)
        except ValueError:
            raise NotFoundError("Couldn't cast the query value to \"{type}\"".format(type=type))

    def get_many(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> Generator[T, None, None]:
        value = query.get(VALUE_KEY)
        count = query.get(COUNT_KEY)

        try:
            # noinspection PyCallingNonCallable
            value = type(value)
        except ValueError:
            raise NotFoundError("Couldn't cast the query value to \"{type}\"".format(type=type))

        return (value for _ in range(count))


class IntFloatDataSource(DataSource):
    @DataSource.dispatch
    def get(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    @get.register(int)
    def get_int(self, query: Mapping[str, Any], context: PipelineContext = None) -> int:
        value = query.get(VALUE_KEY)

        try:
            return int(value)
        except ValueError:
            raise NotFoundError("Couldn't cast the query value to \"int\"")

    @get_many.register(int)
    def get_many_int(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[int, None, None]:
        value = query.get(VALUE_KEY)
        count = query.get(COUNT_KEY)

        try:
            value = int(value)
        except ValueError:
            raise NotFoundError("Couldn't cast the query value to \"int\"")

        return (value for _ in range(count))

    @get.register(float)
    def get_float(self, query: Mapping[str, Any], context: PipelineContext = None) -> float:
        value = query.get(VALUE_KEY)

        try:
            return float(value)
        except ValueError:
            raise NotFoundError("Couldn't cast the query value to \"float\"")

    @get_many.register(float)
    def get_many_float(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[float, None, None]:
        value = query.get(VALUE_KEY)
        count = query.get(COUNT_KEY)

        try:
            value = float(value)
        except ValueError:
            raise NotFoundError("Couldn't cast the query value to \"float\"")

        return (value for _ in range(count))


class StringDataSource(DataSource):
    @DataSource.dispatch
    def get(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    @get.register(str)
    def get_str(self, query: Mapping[str, Any], context: PipelineContext = None) -> str:
        value = query.get(VALUE_KEY)

        try:
            return str(value)
        except ValueError:
            raise NotFoundError("Couldn't cast the query value to \"str\"")

    @get_many.register(str)
    def get_many_str(self, query: Mapping[str, Any], context: PipelineContext = None) -> Generator[str, None, None]:
        value = query.get(VALUE_KEY)
        count = query.get(COUNT_KEY)

        try:
            value = str(value)
        except ValueError:
            raise NotFoundError("Couldn't cast the query value to \"str\"")

        return (value for _ in range(count))


########################
# Unsupported Function #
########################

def test_unsupported():
    from datapipelines import UnsupportedError

    unsupported = DataSource.unsupported(int)
    assert type(unsupported) is UnsupportedError

    unsupported = DataSource.unsupported(float)
    assert type(unsupported) is UnsupportedError

    unsupported = DataSource.unsupported(str)
    assert type(unsupported) is UnsupportedError


#####################
# Provides Function #
#####################

def test_provides():
    source = IntFloatDataSource()

    assert source.provides == {int, float}


def test_wildcard_provides():
    from datapipelines import TYPE_WILDCARD
    source = SimpleWildcardDataSource()

    assert source.provides is TYPE_WILDCARD


################
# Get Function #
################

def test_get():
    source = IntFloatDataSource()

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value}
        result = source.get(int, query)

        assert type(result) is int
        assert result == value

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value}
        result = source.get(float, query)

        assert type(result) is float
        assert result == value


def test_get_unsupported():
    from datapipelines import UnsupportedError
    source = IntFloatDataSource()

    query = {VALUE_KEY: "test"}

    with pytest.raises(UnsupportedError):
        source.get(str, query)


def test_wildcard_get():
    source = SimpleWildcardDataSource()

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value}
        result = source.get(int, query)

        assert type(result) is int
        assert result == value

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value}
        result = source.get(float, query)

        assert type(result) is float
        assert result == value


#####################
# Get Many Function #
#####################

def test_get_many():
    source = IntFloatDataSource()

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value, COUNT_KEY: VALUES_COUNT}
        result = source.get_many(int, query)

        assert type(result) is GENERATOR_CLASS
        for res in result:
            assert type(res) is int
            assert res == value

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value, COUNT_KEY: VALUES_COUNT}
        result = source.get_many(float, query)

        assert type(result) is GENERATOR_CLASS
        for res in result:
            assert type(res) is float
            assert res == value


def test_get_many_unsupported():
    from datapipelines import UnsupportedError
    source = IntFloatDataSource()

    query = {VALUE_KEY: "test", COUNT_KEY: VALUES_COUNT}

    with pytest.raises(UnsupportedError):
        source.get_many(str, query)


def test_wildcard_get_many():
    source = SimpleWildcardDataSource()

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value, COUNT_KEY: VALUES_COUNT}
        result = source.get_many(int, query)

        assert type(result) is GENERATOR_CLASS
        for res in result:
            assert type(res) is int
            assert res == value

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value, COUNT_KEY: VALUES_COUNT}
        result = source.get_many(float, query)

        assert type(result) is GENERATOR_CLASS
        for res in result:
            assert type(res) is float
            assert res == value


#######################
# CompositeDataSource #
#######################

def test_composite_provides():
    source = CompositeDataSource({IntFloatDataSource(), StringDataSource()})
    assert source.provides == {int, float, str}


def test_composite_get():
    source = CompositeDataSource({IntFloatDataSource(), StringDataSource()})

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value}
        result = source.get(int, query)

        assert type(result) is int
        assert result == value

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value}
        result = source.get(float, query)

        assert type(result) is float
        assert result == value

    values = [str(random.uniform(-VALUES_MAX, VALUES_MAX)) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value}
        result = source.get(str, query)

        assert type(result) is str
        assert result == value


def test_composite_get_unsupported():
    from datapipelines import UnsupportedError
    source = CompositeDataSource({IntFloatDataSource(), StringDataSource()})

    query = {VALUE_KEY: bytes()}

    with pytest.raises(UnsupportedError):
        source.get(bytes, query)


def test_composite_get_many():
    source = CompositeDataSource({IntFloatDataSource(), StringDataSource()})

    values = [random.randint(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value, COUNT_KEY: VALUES_COUNT}
        result = source.get_many(int, query)

        assert type(result) is GENERATOR_CLASS
        for res in result:
            assert type(res) is int
            assert res == value

    values = [random.uniform(-VALUES_MAX, VALUES_MAX) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value, COUNT_KEY: VALUES_COUNT}
        result = source.get_many(float, query)

        assert type(result) is GENERATOR_CLASS
        for res in result:
            assert type(res) is float
            assert res == value

    values = [str(random.uniform(-VALUES_MAX, VALUES_MAX)) for _ in range(VALUES_COUNT)]

    for value in values:
        query = {VALUE_KEY: value, COUNT_KEY: VALUES_COUNT}
        result = source.get_many(str, query)

        assert type(result) is GENERATOR_CLASS
        for res in result:
            assert type(res) is str
            assert res == value


def test_composite_get_many_unsupported():
    from datapipelines import UnsupportedError
    source = CompositeDataSource({IntFloatDataSource(), StringDataSource()})

    query = {VALUE_KEY: bytes(), COUNT_KEY: VALUES_COUNT}

    with pytest.raises(UnsupportedError):
        source.get_many(bytes, query)
